/* TroiaPreview
 * version 0.4 date 2004-10-23
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <jpeglib.h>

#include "bmp.h"
#include "Images.h"

//load an image from a bitmap-file
int LoadBitmapFile( char * FileName, long * pSizeX, long * pSizeY, unsigned char * * ppData )
{
  //local variables
  FILE * File;
  long I, X, Y, Pos, DummyCnt;
  BITMAPFILEHEADER BitmapFileHeader;
  BITMAPINFOHEADER BitmapInfoHeader;
  RGBQUAD * pColors;
  int BottomUp;
  unsigned char * pData;
  unsigned char CI;
  
  //initialize return-values
  *pSizeX = -1;
  *pSizeY = -1;
  *ppData = NULL;

  //empty filename ===> fopen will raise assertion, so exit
  if( FileName == NULL || FileName[0] == 0 )
    return 0;

  //open file
  File = fopen( FileName, "rb" );
  if( File == NULL )
    return 0;

  //read bitmap-file-header
  I = fread( &BitmapFileHeader, sizeof( BITMAPFILEHEADER ), 1, File );
  if( I != 1 )
  {
    //close file
    fclose( File );
    return 0;
  }
  //check bitmap-file-header
  if( ((char *)&(BitmapFileHeader.bfType))[0] != 'B' || ((char *)&(BitmapFileHeader.bfType))[1] != 'M' )
  {
    //close file
    fclose( File );
    return 0;
  }

  //read bitmap-info-header
  I = fread( &BitmapInfoHeader, sizeof( BITMAPINFOHEADER ), 1, File );
  if( I != 1 )
  {
    //close file
    fclose( File );
    return 0;
  }
  //check bitmap-info-header
  if( BitmapInfoHeader.biSize != sizeof( BITMAPINFOHEADER ) ||
    BitmapInfoHeader.biPlanes != 1 || //1 plane
    BitmapInfoHeader.biCompression != BI_RGB ) //no compression
  {
    //close file
    fclose( File );
    return 0;
  }
  //8 bpp bitmap
  if( BitmapInfoHeader.biBitCount == 8 )
  {
    //check number of used colors
    if( BitmapInfoHeader.biClrUsed == 0 )
      BitmapInfoHeader.biClrUsed = 256;
    if( BitmapInfoHeader.biClrUsed > 256 )
    {
      //close file
      fclose( File );
      return 0;
    }
  }
  //24 bpp bitmap
  else if( BitmapInfoHeader.biBitCount == 24 )
  {
    //do not use colors in color-table
    BitmapInfoHeader.biClrUsed = 0;
  }
  //other bitmap-formats than 8 bpp or 24 bpp not supported
  else
  {
    //close file
    fclose( File );
    return 0;
  }

  //load color-table
  pColors = (RGBQUAD *)calloc( BitmapInfoHeader.biClrUsed, sizeof( RGBQUAD ) );
  I = fread( pColors, sizeof( RGBQUAD ), BitmapInfoHeader.biClrUsed, File );
  if( I != (long)BitmapInfoHeader.biClrUsed )
  {
    //free color-table
    free( pColors );
    //close file
    fclose( File );
    return 0;
  }

  //bottom-up-bitmap
  if( BitmapInfoHeader.biHeight < 0 )
  {
    BitmapInfoHeader.biHeight *= -1;
    BottomUp = 0;
  }
  else
  {
    BottomUp = 1;
  }

  //go to start of pixels
  if( fseek( File, BitmapFileHeader.bfOffBits, SEEK_SET ) != 0 )
  {
    //free color-table
    free( pColors );
    //close file
    fclose( File );
    return 0;
  }

  //allocate buffer for pixels
  pData = (unsigned char *)malloc( BitmapInfoHeader.biHeight * ((BitmapInfoHeader.biWidth * 3 + 3) & ~3) );

  //calculate number of dummy-characters between lines (lines are 4-byte-aligned)
  if( BitmapInfoHeader.biBitCount == 8 ) //8 bpp
    DummyCnt = 3 - ((BitmapInfoHeader.biWidth + 3) & 0x00000003); 
  else //24 bpp
    DummyCnt = 3 - ((BitmapInfoHeader.biWidth * 3 + 3) & 0x00000003);
  
  //read image
  for( Y = 0; Y < BitmapInfoHeader.biHeight; Y++ )
  {
    //read line
    for( X = 0; X < BitmapInfoHeader.biWidth; X++ )
    {
      //get position in buffer
      if( BottomUp )
        Pos = (BitmapInfoHeader.biHeight - Y - 1) * ((BitmapInfoHeader.biWidth * 3 + 3) & ~3) + X * 3;
      else
        Pos = Y * ((BitmapInfoHeader.biWidth * 3 + 3) & ~3) + X * 3;
      //read pixel 8 bpp
      if( BitmapInfoHeader.biBitCount == 8 )
      {
        //read color-index
        if( fread( &CI, 1, 1, File ) != 1 )
          break;
        //check color
        if( CI >= BitmapInfoHeader.biClrUsed )
          break;
        //get color from color-table
        memcpy( pData + Pos, pColors + CI, 3 );
      }
      //read pixel 24 bpp
      else
      {
        //read color
        if( fread( pData + Pos, 1, 3, File ) != 3 )
          break;
      }
    }
    if( X < BitmapInfoHeader.biWidth ) //error in for-loop (X)
      break;
    //go to next 4-byte-boundary
    if( fseek( File, DummyCnt, SEEK_CUR ) != 0 )
      break;
  }
  if( Y < BitmapInfoHeader.biHeight ) //error in for-loop (Y)
  {
    //free bitmap-data
    free( pData );
    //free color-table
    free( pColors );
    //close file
    fclose( File );
    return 0;
  }

  //free color-table
  free( pColors );
  //close file
  fclose( File );

  //success
  *pSizeX = BitmapInfoHeader.biWidth;
  *pSizeY = BitmapInfoHeader.biHeight;
  *ppData = pData;
  return 1;
}

//load an image from a jpeg-file
int LoadJpegFile( char * FileName, long * pSizeX, long * pSizeY, unsigned char * * ppData )
{
  //local variables
  FILE * File;
  struct jpeg_error_mgr JpegErr;
  struct jpeg_decompress_struct JpegDecInfo;
  long LineSize, X, Y;
  unsigned char * pData, * pLine, * pPixel, Color;
  
  //initialize return-values
  *pSizeX = -1;
  *pSizeY = -1;
  *ppData = NULL;

  //empty filename ===> fopen will raise assertion, so exit
  if( FileName == NULL || FileName[0] == 0 )
    return 0;

  //open file
  File = fopen( FileName, "rb" );
  if( File == NULL )
    return 0;

  //initialize jpeg-error
  JpegDecInfo.err = jpeg_std_error( &JpegErr );
  //create jpeg-decompression-object
  jpeg_create_decompress( &JpegDecInfo );
  //setup file-handle from which to read
  jpeg_stdio_src( &JpegDecInfo, File );

  //get infos from file
  jpeg_read_header( &JpegDecInfo, 1 );
  //set output-format to 3 bytes per pixel and RGB
  JpegDecInfo.output_components = 3;
  JpegDecInfo.out_color_space = JCS_RGB;
  //start decompressing image
  jpeg_start_decompress( &JpegDecInfo );

  //allocate buffer for pixels
  LineSize = ((JpegDecInfo.output_width * 3 + 3) & ~3);
  pData = (unsigned char *)malloc( JpegDecInfo.output_height * LineSize );

  //read all scanlines of image and write them into buffer
  while( JpegDecInfo.output_scanline < JpegDecInfo.output_height )
  {
    pLine = pData + JpegDecInfo.output_scanline * LineSize;
    jpeg_read_scanlines( &JpegDecInfo, &pLine, 1 );
  }

  //finish decompression
  jpeg_finish_decompress( &JpegDecInfo );
  //destroy jpeg-decompression-object
  jpeg_destroy_decompress( &JpegDecInfo );

  //close file
  fclose( File );

  //convert from RGB to BGR
  pLine = pData; //first line
  for( Y = 0; Y < (long)JpegDecInfo.output_height; Y++ )
  {
    pPixel = pLine; //first pixel of line
    for( X = 0; X < (long)JpegDecInfo.output_width; X++ )
    {
      Color = pPixel[0]; //exchange red and blue
      pPixel[0] = pPixel[2];
      pPixel[2] = Color;
      pPixel += 3; //next pixel
    }
    pLine += LineSize; //next line
  }

  //success
  *pSizeX = JpegDecInfo.output_width;
  *pSizeY = JpegDecInfo.output_height;
  *ppData = pData;
  return 1;
}

//resize an image (old image stays unaltered)
void ImageResize( long OldSizeX, long OldSizeY, unsigned char * pOldData, long NewSizeX, long NewSizeY, unsigned char * * ppNewData )
{
  //local variables
  long NewX, NewY, FirstX, LastX, FirstY, LastY, OldX, OldY, Pos;
  double FactorX, FactorY, OldX1, OldX2, OldY1, OldY2;
  double FirstXAntiFactor, LastXAntiFactor, FirstYAntiFactor, LastYAntiFactor;
  double LineFactor, PixelFactor, SumB, SumG, SumR;
  unsigned char * pNewData;

  //allocate buffer for new image
  pNewData = (unsigned char *)malloc( NewSizeY * ((NewSizeX * 3 + 3) & ~3) );

  //calculate factors
  FactorX = (double)OldSizeX / (double)NewSizeX;
  FactorY = (double)OldSizeY / (double)NewSizeY;
  
  //loop through new image
  for( NewY = 0; NewY < NewSizeY; NewY++ )
  for( NewX = 0; NewX < NewSizeX; NewX++ )
  {
    //the new pixel is (NewX, NewY)-(NewX+1, NewY+1)
    //calculate the coordinates of the old pixels
    OldX1 = (double)NewX * FactorX;
    OldX2 = (double)(NewX+1) * FactorX;
    OldY1 = (double)NewY * FactorY;
    OldY2 = (double)(NewY+1) * FactorY;
    //calculate first and last coordinates and factors
    FirstX = (long)OldX1;
    FirstXAntiFactor = OldX1 - (double)FirstX;
    LastX = (long)OldX2;
    LastXAntiFactor = 1.0 - (OldX2 - (double)LastX);
    FirstY = (long)OldY1;
    FirstYAntiFactor = OldY1 - (double)FirstY;
    LastY = (long)OldY2;
    LastYAntiFactor = 1.0 - (OldY2 - (double)LastY);
    //get pixel-color-sums
    SumB = 0;
    SumG = 0;
    SumR = 0;
    //loop through rectangle
    for( OldY = FirstY; OldY <= LastY; OldY++ )
    {
      //get line-factor
      LineFactor = 1.0;
      if( OldY == FirstY )
        LineFactor -= FirstYAntiFactor;
      if( OldY == LastY )
        LineFactor -= LastYAntiFactor;
      //loop through line
      for( OldX = FirstX; OldX <= LastX; OldX++ )
      {
        //get pixel factor
        PixelFactor = 1.0;
        if( OldX == FirstX )
          PixelFactor -= FirstXAntiFactor;
        if( OldX == LastX )
          PixelFactor -= LastXAntiFactor;
        PixelFactor *= LineFactor;
        //get color-values
        Pos = OldY * ((OldSizeX * 3 + 3) & ~3) + OldX * 3;
        SumB += (double)pOldData[Pos + 0] * PixelFactor;
        SumG += (double)pOldData[Pos + 1] * PixelFactor;
        SumR += (double)pOldData[Pos + 2] * PixelFactor;
      }
    }
    //set color-values
    PixelFactor = (OldX2 - OldX1) * (OldY2 - OldY1);
    Pos = NewY * ((NewSizeX * 3 + 3) & ~3) + NewX * 3;
    pNewData[Pos + 0] = (unsigned char)(SumB / PixelFactor);
    pNewData[Pos + 1] = (unsigned char)(SumG / PixelFactor);
    pNewData[Pos + 2] = (unsigned char)(SumR / PixelFactor);
  }
  
  //return new image
  *ppNewData = pNewData;
}

//load an image and resize
int LoadImageResize( char * FileName, long ReSizeX, long ReSizeY, unsigned char * * ppRetData,
  int (* LoadImageFileFunction)( char * FileName, long * pSizeX, long * pSizeY, unsigned char * * ppData ) )
{
  //local variables
  long SizeX, SizeY;
  unsigned char * pData, * pResizeData;

  //load image
  if( !LoadImageFileFunction( FileName, &SizeX, &SizeY, &pData ) )
    //return error
    return 0;

  //size is correct
  if( SizeX == ReSizeX && SizeY == ReSizeY )
  {
    //return original image
    *ppRetData = pData;
    return 1;
  }

  //resize image
  ImageResize( SizeX, SizeY, pData, ReSizeX, ReSizeY, &pResizeData );

  //free original image
  free( pData );

  //return resized image
  *ppRetData = pResizeData;
  return 1;
}
  
//load an image in pieces
//total size: (SizeX * CntX) x (SizeY * CntY)
//size of one piece: SizeX x SizeY
//pPieceData: array of unsigend char * to receive pointers to data of pieces
int LoadImagePieces( char * FileName, long SizeX, long SizeY, long CntX, long CntY, unsigned char * pPieceData[],
  int (* LoadImageFileFunction)( char * FileName, long * pSizeX, long * pSizeY, unsigned char * * ppData ) )
{
  //local variables
  unsigned char * pData;
  long X, Y, YY;

  //load bitmap
  if( !LoadImageResize( FileName, SizeX * CntX, SizeY * CntY, &pData, LoadImageFileFunction ) )
    return 0;

  //allocate memory for pieces
  for( Y = 0; Y < CntY; Y++ )
  for( X = 0; X < CntX; X++ )
    pPieceData[Y*CntX+X] = (unsigned char *)malloc( SizeY * ((SizeX * 3 + 3) & ~3) );

  //cut data into pieces
  for( Y = 0; Y < CntY; Y++ )
  for( X = 0; X < CntX; X++ )
  {
    for( YY = 0; YY < SizeY; YY++ )
      memcpy( pPieceData[Y*CntX+X] + YY * ((SizeX * 3 + 3) & ~3),
        pData + (Y * SizeY + YY) * ((SizeX * CntX * 3 + 3) & ~3) + X * SizeX * 3,
        SizeX * 3 );
  }

  //free original bitmap-data
  free( pData );

  //return success
  return 1;
}
